# Copyright (c) 2016-2024 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from typing import Dict, Optional, final

from lucupy.meta import Singleton

from lucupy.minimodel.resource import Resource
from lucupy.minimodel.resource_type import ResourceType

_all__ = [
    'ResourceManager',
]


@final
class ResourceManager(metaclass=Singleton):
    """
    A singleton class that manages Resource instances to reuse them as per the flyweight design pattern.
    """
    def __init__(self):
        """
        Create an empty dictionary of mappings from name to Resource.
        """
        self._all_resources: Dict[str, Resource] = {}

    def lookup_resource(self,
                        rid: str,
                        description: Optional[str] = None,
                        rtype: Optional[ResourceType] = ResourceType.NONE) -> Optional[Resource]:
        """
        Function to perform Resource caching and minimize the number of Resource objects by attempting to reuse
        Resource objects with the same ID.

        If resource_id evaluates to False, return None.
        Otherwise, check if a Resource with id already exists.
        If it does, return it.
        If not, create it, add it to the map of all Resources, and then return it.

        Note that even if multiple objects do exist with the same ID, they will be considered equal by the
        Resource equality comparator.
        """
        # The Resource constructor raises an exception for id None or containing any capitalization of "none".
        if not rid:
            return None
        if rid not in self._all_resources:
            self._all_resources[rid] = Resource(id=rid, description=description,
                                                type=rtype, legal_creation=True)
        return self._all_resources[rid]



# Copyright (c) 2016-2024 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from .resource_manager import *



# Copyright (c) 2016-2024 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum, IntEnum, auto
from typing import ClassVar, FrozenSet, List, Optional, final

import numpy as np

from lucupy.decorators import immutable
from lucupy.types import ZeroTime
from lucupy.minimodel.obs_filter import obs_is_not_inactive, obs_is_science_or_progcal

from .group import ROOT_GROUP_ID, Group, Group
from .ids import ObservationID, ProgramID, UniqueGroupID
from .observation import Observation, Priority
from .semester import Semester
from .timeallocation import TimeAllocation, TimeUsed, Band
from .too import TooType

__all__ = [
    # 'Band',
    'Program',
    'ProgramMode',
    'ProgramType',
    'ProgramTypes',
]


@final
class ProgramMode(IntEnum):
    """
    Main operational mode, which is one of:
    * Queue
    * Classical
    * Priority Visitor (hybrid mode between queue and classical)
    """
    QUEUE = auto()
    CLASSICAL = auto()
    PV = auto()


@final
@immutable
@dataclass(frozen=True)
class ProgramType:
    """
    Represents the information encompassing the type of program.
    * abbreviation: the code used by the program type (e.g. Q, C, FT, LP)
    * name: user readable representation of the program type
    * is_science: indicates if this program type is a science program

    NOTE that ProgramType instances should NEVER be explicitly created.
    All of the valid ProgramType instances are contained in the ProgramTypes enum
    and should be accessed from there.
    """
    abbreviation: str
    name: str
    is_science: bool = True


@final
class ProgramTypes(Enum):
    """
    A complete list of the ProgramType instances used by Gemini.
    As mentioned in ProgramType, ProgramType should never be instantiated
    outside of this enum: instead, ProgramType instances should be retrieved
    from here.
    """
    C = ProgramType('C', 'Classical')
    CAL = ProgramType('CAL', 'Calibration', False)
    DD = ProgramType('DD', "Director's Time")
    DS = ProgramType('DS', 'Demo Science')
    ENG = ProgramType('ENG', 'Engineering', False)
    FT = ProgramType('FT', 'Fast Turnaround')
    LP = ProgramType('LP', 'Large Program')
    Q = ProgramType('Q', 'Queue')
    SV = ProgramType('SV', 'System Verification')


@final
@dataclass
class Program:
    """
    Representation of a program.

    The FUZZY_BOUNDARY is a constant that allows for a fuzzy boundary for a program's
    start and end times.
    """
    id: ProgramID
    internal_id: str
    # Some programs do not have a typical name and thus cannot be associated with a semester.
    semester: Optional[Semester]
    thesis: bool
    mode: ProgramMode
    type: Optional[ProgramTypes]
    start: datetime
    end: datetime
    allocated_time: FrozenSet[TimeAllocation] = field(hash=False, compare=False)
    used_time: FrozenSet[TimeUsed] = field(hash=False, compare=False)

    # Root group is immutable and should not be used in hashing or comparisons.
    root_group: Group

    band: Optional[Band] = None
    too_type: Optional[TooType] = None

    FUZZY_BOUNDARY: ClassVar[timedelta] = timedelta(days=14)

    def __post_init__(self):
        if self.root_group.id != ROOT_GROUP_ID:
            raise ValueError(f"Program {self.id.id} should have root group should named {ROOT_GROUP_ID.id}, received: "
                             f'"{self.root_group.id}".')

    def bands(self) -> List[Band]:
        """Unique list of ranking bands, the bands must be included in the allocated_time"""
        seen = set()
        seen_add = seen.add
        return [t.band for t in self.allocated_time if not (t.band in seen or seen_add(t.band))]

    def program_awarded(self, band: Band=None) -> timedelta:
        return sum((t.program_awarded for t in self.allocated_time if band is None or t.band == band), ZeroTime)

    def program_previously_used(self, band: Band=None) -> timedelta:
        return sum((t.program_used for t in self.used_time if band is None or t.band == band), ZeroTime)

    def partner_awarded(self, band: Band=None) -> timedelta:
        return sum((t.partner_awarded for t in self.allocated_time if band is None or t.band == band), ZeroTime)

    def partner_previously_used(self, band: Band=None) -> timedelta:
        return sum((t.partner_used for t in self.used_time if band is None or t.band == band), ZeroTime)

    def total_awarded(self, band: Band=None) -> timedelta:
        return sum((t.total_awarded() for t in self.allocated_time if band is None or t.band == band), ZeroTime)

    def total_previously_used(self, band: Band=None) -> timedelta:
        return sum((t.total_used() for t in self.used_time if band is None or t.band == band), ZeroTime)

    def program_used(self, band: Band=None) -> timedelta:
        return self.root_group.program_used(band) + self.program_previously_used(band)

    def partner_used(self, band: Band=None) -> timedelta:
        return self.root_group.partner_used(band) + self.partner_previously_used(band)

    def total_used(self, band: Band=None) -> timedelta:
        return self.root_group.total_used(band) + self.total_previously_used(band)

    def not_charged(self) -> timedelta:
        return self.root_group.not_charged() + sum((t.not_charged for t in self.used_time), ZeroTime)

    def observations(self) -> List[Observation]:
        return self.root_group.observations()

    def get_group(self, group_id: UniqueGroupID) -> Optional[Group]:
        """
        Given a UniqueGroupID, find the subgroup with the ID if it exists.
        Returns None if no such group can be found.
        """
        def aux(group: Group) -> Optional[Group]:
            if group.unique_id == group_id:
                return group
            elif not isinstance(group.children, Observation):
                for subgroup in group.children:
                    retval = aux(subgroup)
                    if retval is not None:
                        return retval
            return None

        return aux(self.root_group)

    def get_observation(self, observation_id: ObservationID) -> Optional[Observation]:
        """
        Given an ObservationID, find the Observation with the ID if it exists.
        Returns None if no such Observation can be found.
        """
        def aux(group: Group) -> Optional[Observation]:
            match group.children:
                case Observation():
                    if group.children.id == observation_id:
                        return group.children
                    else:
                        return None
                case _:
                    for subgroup in group.children:
                        check_subgroup = aux(subgroup)
                        if check_subgroup is not None:
                            return check_subgroup
                    return None
        return aux(self.root_group)

    def mean_priority(self) -> float:
        """
        Return the mean user priority from the active SCIENCE and PROGCAL observations.
        This will be a value in the interval [Priority.LOW.value, Priority.HIGH.value] indicating
        the mean over Observations.

        :return: float
        """
        priorities = [obs.priority.value for obs in self.observations()
                      if obs_is_science_or_progcal(obs) and obs_is_not_inactive(obs)]
        return np.mean(priorities) if len(priorities) else Priority.LOW.value

    def show(self):
        """Print content of the Program.
        """
        print(f'Program: {self.id.id}')
        # Print the group and atom information.
        self.root_group.show(1)



# Copyright (c) 2016-2024 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause
from typing import Any, Dict, Type

__all__ = [
    'Singleton',
]


class Singleton(type):
    """
    Thread-safe Singleton metaclass.
    """
    _instances: Dict[Type[Any], Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in Singleton._instances:
            Singleton._instances[cls] = super().__call__(*args, **kwargs)
        return Singleton._instances[cls]



# Copyright (c) 2016-2024 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from .observation import Observation, ObservationClass, ObservationStatus


__all__ = [
    'obs_is_science_or_progcal',
    'obs_is_not_inactive',
]


_OBS_CLASSES = frozenset({ObservationClass.SCIENCE, ObservationClass.PROGCAL})


def obs_is_science_or_progcal(obs: Observation) -> bool:
    """
    Return True if the Observation is a SCIENCE or PROGCAL observation.
    :param obs: the Observation to check
    :return: True if the Observation is a SCIENCE or PROGCAL observation and False otherwise
    """
    return obs.obs_class in _OBS_CLASSES


def obs_is_not_inactive(obs: Observation) -> bool:
    """
    Return True if the Observation is not INACTIVE.
    :param obs: the Observation to check
    :return: True if Observation is not INACTIVE and False otherwise
    """
    return obs.status != ObservationStatus.INACTIVE



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

"""
Constants that are used here and there.  Some are Quantities,
others are just floats. Not all are used.

The planet-coefs are for series expansions for the phase functions
of the planets, used in predicting apparent magnitude. See code.
"""

from math import pi
from typing import Final

import astropy.units as u
from astropy.coordinates.distances import Distance
from astropy.time import Time

__all__ = [
    'PI',
    'TWOPI',
    'PI_OVER_2',
    'ARCSEC_IN_RADIAN',
    'DEG_IN_RADIAN',
    'HRS_IN_RADIAN',
    'KMS_AUDAY',
    'SPEED_OF_LIGHT',
    'SS_MASS',
    'J2000',
    'J2000_Time',
    'JYEAR',
    'JYEAR_100',
    'SEC_IN_DAY',
    'FLATTEN',
    'EQUAT_RAD',
    'EARTHRAD_IN_AU',
    'ASTRO_UNIT',
    'RSUN',
    'RMOON',
    'PLANET_TOL',
    'KZEN',
]

PI: Final[float] = pi
TWOPI: Final[float] = 2 * PI
PI_OVER_2: Final[float] = PI / 2
ARCSEC_IN_RADIAN: Final[float] = 360 * 60 * 60 / (2 * PI)
DEG_IN_RADIAN: Final[float] = 180 / PI
HRS_IN_RADIAN: Final[float] = 24 / (2 * PI)
KMS_AUDAY: Final[float] = 1731.45683633  # km per sec in 1 AU/day
SPEED_OF_LIGHT: Final[float] = 299792.458  # in km per sec ... exact.
SS_MASS: Final[float] = 1.00134198  # solar system mass in solar units
J2000: Final[float] = 2451545.  # Julian date at standard epoch
J2000_Time: Final[Time] = Time(2451545., format='jd')  # J2000 rendered as a Time
JYEAR: Final[float] = 365.25 # noqa
JYEAR_100: Final[float] = JYEAR * 100 # noqa
SEC_IN_DAY: Final[float] = 86400.
FLATTEN: Final[float] = 0.003352813  # flattening of earth, 1/298.257
EQUAT_RAD: Final[Distance] = 6378137. * u.m  # equatorial radius of earth, meters
EARTHRAD_IN_AU: Final[float] = 23454.7910556298  # number of earth rad in 1 au
ASTRO_UNIT: Final[float] = 1.4959787066e11  # 1 AU in meters
RSUN: Final[float] = 6.96000e8  # IAU 1976 recom. solar radius, meters
RMOON: Final[float] = 1.738e6  # IAU 1976 recom. lunar radius, meters
PLANET_TOL: Final[float] = 3.  # flag if nearer than 3 degrees
KZEN: Final[float] = 0.172  # mag / airmass relation for Hale Pohaku
